#ifndef _MCDCONVEXMESH_H
#define _MCDCONVEXMESH_H
/* -*- mode: C; -*- */

/*
   Copyright (c) 1997-2002 MathEngine PLC

   $Name:  $

   Date: $Date: 2003/05/05 00:57:07 $ - Revision: $Revision: 1.1 $

   This software and its accompanying manuals have been developed
   by MathEngine PLC ("MathEngine") and the copyright and all other
   intellectual property rights in them belong to MathEngine. All
   rights conferred by law (including rights under international
   copyright conventions) are reserved to MathEngine. This software
   may also incorporate information which is confidential to
   MathEngine.

   Save to the extent permitted by law, or as otherwise expressly
   permitted by MathEngine, this software and the manuals must not
   be copied (in whole or in part), re-arranged, altered or adapted
   in any way without the prior written consent of the Company. In
   addition, the information contained in the software may not be
   disseminated without the prior written consent of MathEngine.

 */


#include <McdGeometry.h>
#include <McdGeometryTypes.h>
#include <McdCTypes.h>
#include <McdInteractionTable.h>
#include <MeInline.h>

#ifdef __cplusplus
extern "C" {
#endif

/** McdConvexMesh is a concrete McdGeometry type. */

MCD_DECLARE_GEOMETRY_TYPE(McdConvexMesh);

/* creation */

MEPUBLIC
McdConvexMeshID   MEAPI McdConvexMeshCreate(McdFramework *frame,  McdConvexHull *poly, 
                                            MeReal fatness);

MEPUBLIC
McdConvexMeshID   MEAPI McdConvexMeshInit(McdConvexMesh *convex, McdFramework *frame, 
                                          McdConvexHull *poly, MeReal fatness);

MEPUBLIC
void              MEAPI McdConvexMeshTerm(McdConvexMeshID convex);

MEPUBLIC
void              MEAPI McdConvexMeshDestroy(McdConvexMeshID convex);

MEPUBLIC
void              MEAPI McdConvexMeshDestroy(McdConvexMeshID);

/**
    Create new convex polyhedron object from its vertices. Computes
    the convex hull polygons and edges.
*/

MEPUBLIC
void              MEAPI McdConvexMeshGetXYAABB( McdConvexMeshID, MeMatrix4 tm,
                                                MeReal bounds[4]);

/**
  This allocates memory for the convex hull based on the sizes.
  One extra face and vertex is required for the sentinel.
*/
void McdAllocateHull(McdConvexHull *cnv);

/**
  This deallocates the memory that is allocated by McdAllocateHull
  or by McdComputeHull.
*/

MEPUBLIC
void McdDeallocateHull(McdConvexHull *cnv);

/**
  This makes a deep copy of a convex hull.  The data in to_cnv is allocated
  with McdAllocateHull().
*/
void McdConvexHullCopy(McdConvexHull *to_cnv, const McdConvexHull *from_cnv);

/**
  This computes the dual of a convex hull.  This is useful for computing
  a convex hull when you only know the planes (for example the shape is a
  BSP node).
*/
void McdConvexHullDual(McdConvexHull *dual, const McdConvexHull *src,
                       MeVector3 offset, MeReal scale);

/**
  This is an internal function that is used to search for an edge.
    f is the face (or -1 for any face)
    to is 1=to, 0=from
    v is the vertex index
*/
int McdFindEdge(const McdConvexHull *cnv, int f, int to, int v);

/**
  This precomputes some stuff, like the invLength of each edge.
*/
void McdPrecomputeStuff(McdConvexHull *cnv);

/**
  This populates a triangle convex hull.  It must have 3 vert, 1 face, and
  3 edges.  The points p1,p2,p3 must be ACW.
*/
void McdGetTriangleHull(McdConvexHull *cnv, const MeVector3 p1, 
                        const MeVector3 p2, const MeVector3 p3);

/*
  ACCESSOR METHODS

  Functions are provided below for convenience to access common things
  that are not trivial to access directly in the data structures.
*/

/*
    =======================================================================
        ACCESSOR METHODS
    =======================================================================
*/

/**
  Get the number of edges (or vertices or adjacent faces) around a given face.
*/
MeINLINE
int McdCnvFaceGetCount(const McdConvexHull *cnv, int face)
{
    MEASSERT(cnv && face>=0 && face < cnv->numFace);

    return cnv->firstEdge[face+1] - cnv->firstEdge[face];
}

/**
  Get the i-th edge of a face.  The edges are guaranteed to form a directed 
  path (from-to) in ACW order around the face. 
  You can look at the edge.rightFace to get all adjacent faces.
*/
MeINLINE
const McdCnvEdge *McdCnvFaceGetEdge(const McdConvexHull *cnv, int face, int i)
{
    MEASSERT(i >= 0 && i < McdCnvFaceGetCount(cnv, face));

    return cnv->edge + (i + cnv->firstEdge[face]);
}

/**
  These functions get the i-th vertex on the face.
*/
MeINLINE
int McdCnvFaceGetVertexId(const McdConvexHull *cnv, int face, int i)
{
    return McdCnvFaceGetEdge(cnv, face, i)->fromVert;
}

MeINLINE
MeVector3Ptr McdCnvFaceGetVertex(const McdConvexHull *cnv,
    int face, int i)
{
    return cnv->vertex[McdCnvFaceGetVertexId(cnv, face, i)];
}

/**
  Get the number of edges (or vertices or faces) around a given vertex.
*/
MeINLINE
int McdCnvVertexGetCount(const McdConvexHull *cnv, int vertex)
{
    MEASSERT(cnv && vertex>=0 && vertex < cnv->numVertex);

    return cnv->firstEdgeIndex[vertex+1] - cnv->firstEdgeIndex[vertex];
}

/**
  Get the id of the i-th edge around a vertex.
*/
MeINLINE
int McdCnvVertexGetEdgeId(const McdConvexHull *cnv, int vertex, int i)
{
    MEASSERT(i >= 0 && i < McdCnvVertexGetCount(cnv, vertex));

    return cnv->edgeIndex[i + cnv->firstEdgeIndex[vertex]];
}

/**
  Get the i-th edge from a vertex.    The edges are guaranteed
  to be all "from" the specified vertex "to" another vertex.
  You can look at the edge.rightFace to get all adjacent faces.
*/
MeINLINE
const McdCnvEdge *McdCnvVertexGetEdge(const McdConvexHull *cnv, int vertex, int i)
{
    return cnv->edge + McdCnvVertexGetEdgeId(cnv, vertex, i);
}

/**
  Get the id of the i-th vertex that is adjacent to a given vertex.
*/
MeINLINE
int McdCnvVertexGetNeighbor(const McdConvexHull *cnv, int vertex, int i)
{
    return cnv->edge[McdCnvVertexGetEdgeId(cnv, vertex, i)].toVert;
}



/* accessors */

MEPUBLIC
int               MEAPI McdConvexMeshGetPolygonCount(McdConvexMeshID );
MEPUBLIC
int               MEAPI McdConvexMeshGetPolygonVertexCount(McdConvexMeshID,
                            int polyID);
MEPUBLIC
void              MEAPI McdConvexMeshGetPolygonVertex(McdConvexMeshID,
                            int polyID, int vertexID, MeVector3);
MEPUBLIC
MeVector3ConstPtr MEAPI McdConvexMeshGetPolygonVertexPtr(McdConvexMeshID,
                            int polyID, int vertexID);
MEPUBLIC
void              MEAPI McdConvexMeshGetPolygonNormal(McdConvexMeshID, int,
                            MeVector3);
MEPUBLIC
MeI16             MEAPI McdConvexMeshGetMassProperties( McdConvexMeshID mesh, MeMatrix4 relTM,
                            MeMatrix3 m, MeReal* volume);

MEPUBLIC
MeReal            MEAPI McdConvexMeshGetFatness(McdConvexMeshID mesh);

MEPUBLIC
const McdConvexHull * MEAPI McdConvexMeshGetPolyhedron(McdConvexMeshID);
MEPUBLIC
void              MEAPI McdConvexMeshSetPolyhedron(McdConvexMeshID, McdConvexHull *poly, MeReal fatness);

  /* interactions */

MEPUBLIC
MeBool            MEAPI McdConvexMeshRegisterBoxAndSphereFns(McdFramework *frame);
MEPUBLIC
MeBool            MEAPI McdConvexMeshPrimitivesRegisterInteractions(McdFramework *frame);

MCD_DECLARE_INTERSECT_INTERACTION(ConvexMesh, ConvexMesh);
MCD_DECLARE_INTERSECT_INTERACTION(ConvexMesh, Plane);
MCD_DECLARE_INTERSECT_INTERACTION(Sphyl, ConvexMesh);
MCD_DECLARE_INTERSECT_INTERACTION(Sphere, ConvexMesh);
MCD_DECLARE_INTERSECT_INTERACTION(Box, ConvexMesh);
MCD_DECLARE_INTERSECT_INTERACTION(Plane, ConvexMesh);
MCD_DECLARE_SAFETIME_INTERACTION(ConvexMesh, TriangleList);
MCD_DECLARE_INTERSECT_INTERACTION(Cylinder, ConvexMesh);
MCD_DECLARE_INTERSECT_INTERACTION(ConvexMesh, RGHeightField);

MCD_DECLARE_LINESEG_INTERACTION(ConvexMesh);

#ifdef __cplusplus
}
#endif

#endif /* _MCDCONVEXMESH_H */
