#ifndef _MCDGEOMETRYTYPES_H /* -*- mode: C; -*- */
#define _MCDGEOMETRYTYPES_H

/*
   Copyright (c) 1997-2002 MathEngine PLC

   $Name:  $

   Date: $Date: 2003/05/05 00:57:07 $ - Revision: $Revision: 1.1 $

   This software and its accompanying manuals have been developed
   by MathEngine PLC ("MathEngine") and the copyright and all other
   intellectual property rights in them belong to MathEngine. All
   rights conferred by law (including rights under international
   copyright conventions) are reserved to MathEngine. This software
   may also incorporate information which is confidential to
   MathEngine.

   Save to the extent permitted by law, or as otherwise expressly
   permitted by MathEngine, this software and the manuals must not
   be copied (in whole or in part), re-arranged, altered or adapted
   in any way without the prior written consent of the Company. In
   addition, the information contained in the software may not be
   disseminated without the prior written consent of MathEngine.
*/

/**
  @file
  All geometry types
*/

#include <McdGeometry.h>

/** Geometry ID's */
/**@{*/
typedef McdGeometryID McdNullID;
typedef McdGeometryID McdSphereID;
typedef McdGeometryID McdBoxID;
typedef McdGeometryID McdPlaneID;
typedef McdGeometryID McdCylinderID;
typedef McdGeometryID McdSphylID;
typedef McdGeometryID McdTriangleListID;
typedef McdGeometryID McdConvexMeshID;
typedef McdGeometryID McdAggregateID;
/**@}*/

enum
{
    kMcdGeometryTypeNull = 0,
    kMcdGeometryTypeSphere = 1,
    kMcdGeometryTypeBox = 2,
    kMcdGeometryTypePlane = 3,
    kMcdGeometryTypeCylinder = 4,
    kMcdGeometryTypeSphyl = 5,
    kMcdGeometryTypeTriangleList = 6,
    kMcdGeometryTypeConvexMesh = 7,
    kMcdGeometryTypeAggregate = 8,
    kMcdGeometryBuiltInTypes = 9
};


/** McdNull is a concrete McdGeometry type with no properties. Attempts to
call virtual functions on this geometry will fail.
@see McdGeometry.h */

typedef struct
{
  McdGeometry m_g; /**< m_g */

  MeReal mR[3];    /**< mR */
  MeReal mRadius;  /**< mRadius */

} McdNull;


/** McdBox is a concrete McdGeometry type. All functions taking type
 * McdGeometryID as argument can be called with a McdBoxID argument.
 * @see McdGeometry.h
 */

typedef struct
{
  McdGeometry m_g; /**< m_g */

  MeReal mR[3];    /**< mR */
  MeReal mRadius;  /**< mRadius */

} McdBox;


typedef struct
{
  McdGeometry m_g;

  MeReal m_upperHeight,m_lowerHeight,m_lowerRadius,m_upperRadius,m_sinAngle;

} McdCone;


typedef struct
{
  McdGeometry m_g;

  MeReal mR;
  /* half-height */
  MeReal mRz;
  MeReal mSphereRadius;

} McdCylinder;


/** McdSphere is a concrete McdGeometry type. */
typedef struct
{
  McdGeometry m_g; /**< m_g */

  MeReal mRadius;  /**< radius */

} McdSphere;



typedef struct
{
  McdGeometry m_g;

} McdPlane;


typedef struct
{
    McdGeometry m_g;

    MeReal mRadius;
    MeReal mHalfHeight;
} McdSphyl;

/*@{*/
/** 
 * Collision user triangle. If the triangle is not two-sided, the normal 
 * is assumed to be  facing outwards (i.e. penetration will be along the 
 * direction opposite to the normal). For an edge to be taken into account 
 * when computing the minimal penetration plane, which also defines the 
 * contact normal for any contacts with that triangle, the edge must be set 
 * to be "active". In the standard version of McdTriangleList, all edges are 
 * active, and the triangle is two-sided
*/
typedef enum
{
    kMcdTriangleUseSmallestPenetration = 1,
    kMcdTriangleTwoSided               = 2,
    kMcdTriangleUseEdge0               = 4,  /* use edge v1 - v0 */
    kMcdTriangleUseEdge1               = 8,  /* use edge v2 - v1 */
    kMcdTriangleUseEdge2               = 16, /* use edge v0 - v2 */
    kMcdTriangleUseEdges               = 28,
    kMcdTriangleStandard               = 31
} McdTriangleFlags;


typedef struct _McdUserTriangle
{
    MeVector3 *vertices[3]; /**< pointers to vertices */
    MeVector3 *normal;      /**< normal */
    MeReal distance;        /**< for swept sphere queries only */
    union {void *ptr; int tag;} triangleData; /**< triangleData */
    McdTriangleFlags flags; /**< flags */

} McdUserTriangle;
/*@}*/

/*@{*/
/** 
 * user function to get triangles given a model containing a
 *  triangle list geometry. The triangle list can be either model
 *  in the pair.
 */

typedef int (MEAPI * McdTriangleListFnPtr) (McdModelPair* modelTriListPair,
                                            McdUserTriangle *triangle,
                                            MeVector3 pos, 
                                            MeReal radius,
                                            int maxTriangles);
/** 
 * This is a user function to get triangles that for a swept sphere.
 * The triangles must indicate distance to pos1.  Since the distance is
 * used for culling, it's ok if it is too small, but it should not be
 * larger than actual. 
 * If you do not provide a swept function, Karma will call the
 * regular triangle callback with a large radius and calculate all
 * the distances itself, which may not be most efficient.
 */
typedef int (MEAPI * McdTriangleListSweptFnPtr) 
                                           (McdModelPair* modelTriListPair,
                                            const MeVector3 pos1, 
                                            const MeVector3 pos2, 
                                            MeReal radius);
/*@}*/
/** McdTriangleList */
typedef struct McdTriangleList
{
    McdGeometry m_g; /**< m_g */
    
    MeVector3 center; /**< centre */
    MeVector3 radius; /**< radius */
    int triangleMaxCount; /**< max number of triangles in list */
    void *userData;  /**< user geometry representation, or other user data */

    McdTriangleListFnPtr triangleListGenerator; /**< returns triangles in a
                                                  sphere */    
    McdTriangleListSweptFnPtr triangleListSwept;/**< returns triangles in a
                                                  swept sphere */

    McdUserTriangle *list;  /**< pointer to list. Safe to modify in
                              callback if e.g. you already have the list */
} McdTriangleList;

/** @def MCD_MAX_POLY_VERTICES */
/** Max number of polygon vertices */
#define MCD_MAX_POLY_VERTICES 8

/** User polygon */
typedef struct _McdUserPolygon
{
    int vertexList[MCD_MAX_POLY_VERTICES+1]; /**< pointers to vertices */
    int vertexCount; /**< vertexCount */
    union {void *ptr; int tag;} polyData; /**< polyData */
    int counter; /**< counter */
    int clump; /**< clump */
} McdUserPolygon;

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/*@{*/

/** user function to get triangles given a model containing a
triangle list geometry. The triangle list can be either model
in the pair. */
typedef int (MEAPI * McdPolyMashFnPtr) (McdModelPair* pair,
                                        MeVector3 **vertices,
                                        int *vertexCount,
                                        McdUserPolygon **poly,
                                        int *polyCount);
/*@}*/
#endif
/** McdPolyMash */
typedef struct McdPolyMash
{
    McdGeometry m_g;            /**< m_g */
    MeVector3 centre;           /**< centre */
    MeVector3 radius;           /**< radius */
    McdPolyMashFnPtr generator; /**< generator */
    void *userData;  /**< user geometry representation, or other user data */    
} McdPolyMash;



/**
 * These are the structures used to store a convex hull.
 * 
 * Separate functions are provided to measure/allocate/populate the hull
 * so that you can use any allocation method (such as alloca).
 * 
 * This data allows you to easily traverse the convex hull using a hill 
 * climbing algorithm, either by walking adjacent faces, or walking 
 * adjacent vertices.  Each face points to a list of edges which are sorted
 * in anti-clockwise order around the face.  Also each vertex points to 
 * a list of edges which are sorted ACW around the vertex.  The list of
 * all edges stores every edge two times (one each direction).  Every edge
 * indicates its two end points and the face on its right side.  If you want
 * to look at every edge once only, then you can ignore the backward
 * edges, for instance, fromVert < toVert.
 * 
 * The edges are sorted in anti-clockwise order around each face.
 * To access the edges sorted in anti-clockwise order around a vertex, you go
 * through the "edgeIndex" array.  For example,
 * //  every edge/face/vertex around a vertex in ACW order
 * 
 * for (i=cnv.firstEdgeIndex[v]; i<cnv.firstEdgeIndex[v+1]; ++i)
 * {
 *       cout << cnv.edge[cnv.edgeIndex[i]].rightFace;
 *       cout << cnv.edge[cnv.edgeIndex[i]].toVert;
 *   }
 *
 *   //  every edge/face/vertex around a face in ACW order
 *
 *   for (i=cnv.firstEdge[f]; i<cnv.firstEdge[f+1]; ++i)
 *   {
 *      cout << cnv.edge[i].rightFace;
 *       cout << cnv.edge[i].fromVert;
 *   }
 *
 * For convenience, the last+1 entry in firstEdge (and firstEdgeIndex) 
 * is a sentinel which equals numEdges.
 */
/**@{*/
typedef struct McdConvexHull 
{
    MeVector3 *vertex;       /**< position of all vertices, not in any order */
    int *firstEdgeIndex;     /**< this is used to access edges via vertices */
    MeVector3 *faceNormal;   /**< normals of all faces, not in any order */
    int *firstEdge;          /**< this is used to access edges via faces */
    struct McdCnvEdge *edge; /**< all edges, sorted by leftFace and ACW */
    int *edgeIndex;          /**< this is used to access edges via vertices */
    int numVertex;           /**< numVertex */
    int numFace;             /**< numFace */
    int numEdge;             /**< two times the number of geometric edges */
}
McdConvexHull;

typedef struct McdCnvEdge
{
    MeReal invLength;  /**< 1 / length of edge */
    int fromVert;    /**< index of vertex of start of edge */
    int toVert;      /**< index of vertex of end of edge */
    int rightFace;   /**< index of the face on the right side of the edge */
    int leftFace;    /**< index of the face on the left side of the edge */
}
McdCnvEdge;
/**@}*/

/** McdConvexMesh struct */
typedef struct
{

  McdGeometry m_g;                 /**< m_g      */
  McdConvexHull mHull;             /**< mHull    */
  MeReal mFatness;                 /**< mFatness */
  MeReal mBoundingSphereRadius;    /**< mBoundingSphereRadius */
  MeVector3 mBoundingSphereCenter; /**< mBoundingSphereCenter */
} McdConvexMesh;

/** McdAggregate struct */
typedef struct
{
    McdGeometry m_g;                 /**< m_g */
    MeMatrix4 *tmTable;              /**< tmTable */
    McdGeometryID *geometryTable;    /**< geometryTable */
    int elementCount;                /**< elementCount */
    int elementCountMax;             /**< elementCountMax */
    MeReal mBoundingSphereRadius;    /**< mBoundingSphereRadius */
    MeVector3 mBoundingSphereCenter; /**< mBoundingSphereCenter */
} McdAggregate;

#endif /* _MCDGEOMETRYTYPES_H */
