#ifndef _MCDMODEL_H
#define _MCDMODEL_H
/* -*- mode: C; -*- */

/*
   Copyright (c) 1997-2002 MathEngine PLC

   $Name:  $

   Date: $Date: 2003/05/05 00:57:07 $ - Revision: $Revision: 1.1 $

   This software and its accompanying manuals have been developed
   by MathEngine PLC ("MathEngine") and the copyright and all other
   intellectual property rights in them belong to MathEngine. All
   rights conferred by law (including rights under international
   copyright conventions) are reserved to MathEngine. This software
   may also incorporate information which is confidential to
   MathEngine.

   Save to the extent permitted by law, or as otherwise expressly
   permitted by MathEngine, this software and the manuals must not
   be copied (in whole or in part), re-arranged, altered or adapted
   in any way without the prior written consent of the Company. In
   addition, the information contained in the software may not be
   disseminated without the prior written consent of MathEngine.

 */

/** @file

    The collision model object and utility functions

    An McdModel is a collision model whose geometrical shape is
    specified by a McdGeometry object. An McdModel also holds the
    transformation matrix describing its position and orientiation in 3D
    space.

    Collision models are the principal object type in the Mcd system. Typically 
    there is a one-to-one correspondence between a collision model and 
    a 3D graphics model that is rendered on the screen. Usually there is one
    collision model per dynamics model,

    You can query geometrical properties of an isolated McdModel object
    using the McdModel interface, which uses the McdGeometryInstance protocol 
    and takes into account its transformation matrix. A model may contain a
    tree of geometry instances, in the case that it is an aggregate.

    You can query the geometrical relationship between a pair of McdModel
    objects in close proximity, by creating a McdModelPair object and using
    any of the pairwise queries declared in McdInteractions.h.

    You can insert McdModel objects into McdSpace, which will keep track of
    which pairs are in close proximity.

*/

#include <McdCTypes.h>

#ifdef __cplusplus
extern "C" {
#endif


/**
 * Create a new collision model with geometry @a g.
 *
 * You must explicitly allocate a transformation matrix for the newly
 * created object before it can be used. This is set using either
 * McdModelSetTransformPtr() or McdModelSetRelativeTransformPtrs().
 */

MEPUBLIC
McdModelID        MEAPI McdModelCreate(McdGeometryID g );

/**
 * Destroy @a m. @a m's associated McdGeometry object is not automatically
 * destroyed, since it might be shared by other McdModel objects. It must be
 * destroyed explicitly, if desired.
 */
MEPUBLIC
void              MEAPI McdModelDestroy( McdModelID cm );

/** McdModelInit */ 
MEPUBLIC
McdModelID        MEAPI McdModelInit(McdModelID cm, McdGeometryID g);

/**
 * Finalize @a m. @a m's associated McdGeometry object is not automatically
 * destroyed, since it might be shared by other McdModel objects. It must be
 * destroyed explicitly, if desired.
 */
MEPUBLIC
void              MEAPI McdModelTerm(McdModelID cm);

/**
 * Returns model's collision space. NULL if model is not in a space.
 */
MEPUBLIC
McdSpaceID        MEAPI McdModelGetSpace(McdModelID cm);

/**
 * Returns model's geometry type.
 */
MEPUBLIC
McdGeometryID     MEAPI McdModelGetGeometry( McdModelID g );
/**
 * Sets model's geometry type.
 */
MEPUBLIC
void              MEAPI McdModelSetGeometry( McdModelID cm, McdGeometryID g );

/**
 * Returns the pointer to the model's GeometryInstance structure
 */
MEPUBLIC
McdGeometryInstanceID
                  MEAPI McdModelGetGeometryInstance(McdModelID cm);
/** Returns McdGeometry type */
MEPUBLIC
McdGeometryType   MEAPI McdModelGetGeometryType(McdModelID cm);

/**
 * Sets the geometry's 4X4 transformation matrix.
 *
 * User must allocate his/her own transform for model's geometry,
 * and pass a pointer in with this function. This transform determines
 * the global coordinates of the model's geometry, which is specified
 * locally.
 *
 * If cm has been used in McdModelSetBody(..), then its transform
 * pointer will be pointed to the MdtBody's transform.
 *
 * If McdModelSetTransformPtr() is called after this point, a warning
 * will be issued.
 */
MEPUBLIC
void              MEAPI McdModelSetTransformPtr( const McdModelID cm,
                         const MeMatrix4Ptr geometryTM );
/**
 *  Returns the pointer to the geometry's 4X4 transformation matrix.
 *
 * This transform determines the global coordinates of the
 * model's geometry, which is specified locally.
 *
 * If cm has been used in McdDtBridgeSetBody(..), then its transform
 * pointer will be pointed to the MdtBody's transform.
 *
 * The returned address is intended for reading only, not writing. The
 * validity of the return value is guaranteed only during the current
 * time step.
 */
MEPUBLIC
MeMatrix4Ptr      MEAPI McdModelGetTransformPtr( McdModelID cm );


/**
 * Copies the model's relative transform, allocates space for a compound
 * transform, and then calls McdModelSetRelativeTransformPtrs.
 *
 * The relative and compound matrices are owned by the model and will be
 * deleted when the relative transform pointers are overwritten or when
 * the model is destroyed
 *
 * @see McdModelSetRelativeTransformPtrs.
 */
MEPUBLIC
void              MEAPI McdModelSetRelativeTransform(McdModelID cm,
                                                     MeMatrix4 relTM);

/**
 * Returns a pointer to the model's relative transform 
 * @see McdModelGetRelativeTransformPtrs.
*/
MEPUBLIC
MeMatrix4Ptr      MEAPI McdModelGetRelativeTransform(McdModelID cm);

/**
 * Defines a transform offset between the coordinate system of \a cm and
 * another reference coordinate system.
 * 
 * McdModel always uses the transform set in McdModelSetTransformPtr for
 * collision calculations. To define an offset between an external
 * reference transform and the collision transform, requires both the
 * allocation of \a relativeTM, a relative transform describing the
 * offset value, and \a compoundTM.
 *
 * Each time step, the referenceTM is compounded by \a relativeTM, and
 * the result is written to the transform pointed to by
 * McdModelSetTransformPtr. If \a relativeTM is NULL, then no
 * compounding occurs.
 * 
 * IMPORTANT: This function must be called AFTER McdModelSetBody().
 * @param cm : Model ID
 * @param relTM : the relative transform between the two coordinate
 *   systems. Allocated by user.
 * @param refTM :  The transform representing the external
 *   reference transform.
 * @param compoundTM :  The transform representing the compounded
 *   reference transform.
 * @param own : if true, the transforms will be deleted if a new relative
 *   transform is set. In this case the transform should be allocated using
 *   MeMemoryAPI.createAligned
 *
 * @see McdModelSetTransformPtr
 */
MEPUBLIC
void              MEAPI McdModelSetRelativeTransformPtrs(McdModelID cm,
                                                         MeMatrix4 relTM,
                                                         MeMatrix4 refTM,
                                                         MeMatrix4 compoundTM,
                                                         MeBool own);


/**
 * Set the pointer to the relative transform and to the reference transform.
 *
 * The model's transform (as set by McdModelSetTransformPtr) is updated
 * (in McdModelUpdate) to be the compound of the relative and reference
 * transforms.
 */
MEPUBLIC
void              MEAPI McdModelGetRelativeTransformPtrs(McdModelID cm,
                                                         MeMatrix4Ptr *relTM,
                                                         MeMatrix4Ptr *refTM);

/**
 * Indicates that the transformation, velocities, etc... have changed.
 * Called automatically by McdSpaceUpdate.
 */
MEPUBLIC
void              MEAPI McdModelUpdate( McdModelID cm );
/**
 * Computes and caches the axis-aligned bounding box which contains
 * the model's swept volume as it moves kinematically along the path
 * given by the linear and angular velocities starting from the
 * position given by the transform.
 *
 * This function is subject to change.
*/
MEPUBLIC
void              MEAPI McdModelUpdatePath( McdModelID cm, MeReal motionDuration );
/** To be documented */
MEPUBLIC
void              MEAPI McdModelUpdatePathCompatible( McdModelID cm, MeReal dummy );

MEPUBLIC
void              MEAPI McdModelSetUpdateCallback( McdModelID cm, McdModelUpdateFnPtr f );
MEPUBLIC
McdModelUpdateFnPtr
                  MEAPI McdModelGetUpdateCallback( McdModelID cm );
/**
 * Explicitely updates the model's transform from it's relative and
 * reference transform.
 * @see McdModelGetRelativeTransformPtrs.
 */
MEPUBLIC
void              MEAPI McdModelCompoundTransforms( McdModelID cm );

/**
 * Returns the cached axis-aligned bounding box which contains the
 * model's geometry.
 */
MEPUBLIC
MeBool            MEAPI McdModelGetAABB( McdModelID cm, MeVector3 minCorner,
                            MeVector3 maxCorner );


/**
 * Set a pointer to the models linear velocity vector.
 *
 * This is used in McdModelGetPathAABB and in McdSafeTime.
 */
MEPUBLIC
void              MEAPI McdModelSetLinearVelocityPtr( McdModelID cm, MeReal *);

/**
 * Get the pointer to the models linear velocity vector.
 *
 * See @McdModelSetLinearVelocityPtr
 */
MEPUBLIC
MeReal*           MEAPI McdModelGetLinearVelocityPtr( McdModelID cm );
/**
 * Set a pointer to the models angular velocity vector.
 *
 * This is used in McdModelGetPathAABB and in McdSafeTime.
 */
MEPUBLIC
void              MEAPI McdModelSetAngularVelocityPtr(McdModelID cm, MeReal *);
/**
 * Get the pointer to the models angular velocity vector.
 *
 * See @McdModelSetAngularVelocityPtr
 */
MEPUBLIC
MeReal*           MEAPI McdModelGetAngularVelocityPtr(McdModelID cm );

/**
 * Set a pointer to the models relative center of mass vector.
 *
 * This is used in McdModelGetPathAABB and in McdSafeTime.
 */
MEPUBLIC
void              MEAPI McdModelSetCOMPtr( McdModelID cm, MeReal * com);

/**
 * Get the pointer to the models relative center of mass vector.
 *
 * See @McdModelSetCOMPtr
 */
MEPUBLIC
MeReal*           MEAPI McdModelGetCOMPtr( McdModelID cm );

/**
 * Returns the minimal-volume sphere which completely bounds the volume
 * occupied by @a m.
 */
MEPUBLIC
MeBool            MEAPI McdModelGetBSphere( McdModelID cm, MeVector3 center,
                            MeReal *radius );

/**
 * Sets @a m's contact tolerance to @a tol.
 * Used for geometric intersection tests.
 */
MEPUBLIC
void              MEAPI McdModelSetContactTolerance(McdModelID cm, MeReal tol);

/**
 * Returns @a m's contact tolerance.
 */
MEPUBLIC
MeReal            MEAPI McdModelGetContactTolerance(McdModelID cm );

/**
 * Sets the user data to be attached to the model.
 */
MEPUBLIC
void              MEAPI McdModelSetUserData( McdModelID cm, void *data );
/**
 * Gets the user data attached to the model.
 */
MEPUBLIC
void*             MEAPI McdModelGetUserData( McdModelID cm );

  /** internal */
void*             MEAPI McdModelGetBodyData( McdModelID cm );

  /** internal */
void              MEAPI McdModelSetBodyData( McdModelID cm, void * body );

  /** Set the models material ID. Used in conjunction with Mst. */
MEPUBLIC
void              MEAPI McdModelSetMaterial( McdModelID cm, int material );

/** Read the models material id. Used in conjunction with Mst.
 * The default material ID upon model creation is 0.
 * @see McdModelSetMaterial.
 */
MEPUBLIC
unsigned int      MEAPI McdModelGetMaterial( McdModelID cm );


  /** Set the models request ID. Used in conjunction with McduRequestTable. */
MEPUBLIC
void              MEAPI McdModelSetRequestID( McdModelID cm, int requestId );

  /** Read the models request ID. Used in conjunction with McduRequestTable. */
MEPUBLIC
int               MEAPI McdModelGetRequestID( McdModelID cm );
  /** Get the line segment intersect function associated with the model's geometry */

  /** Read the model's sort key. Used in conjunction with McduRequestTable. */
MEPUBLIC
MeI16             MEAPI McdModelGetSortKey(McdModelID cm);

  /** Set the model's sort key. Should be between 0 and (2^15)-1 */
MEPUBLIC
void              MEAPI McdModelSetSortKey(McdModelID cm, MeI16 key);

/** McdModelSetIntersectCallback */
MEPUBLIC 
void              MEAPI McdModelSetIntersectCallback(McdModelID cm, McdModelIntersectFnPtr fn);

/** McdModelGetIntersectCallback */
MEPUBLIC           
McdModelIntersectFnPtr 
                  MEAPI McdModelGetIntersectCallback(McdModelID cm);

/** McdModelGetDefaultRequestID */
#define McdModelGetDefaultRequestID() (0)

#ifdef __cplusplus
}
#endif

#endif /* _MCDMODEL_H */
