#ifndef MDTSOLVERUTILS_H
#define MDTSOLVERUTILS_H

/**
 * Kea 4x4 Transformation Matrix.
 */
typedef struct
{
    /** Rotation matrix row 0 (last element 0). */
    MeVector4 R0;
    /** Rotation matrix row 1 (last element 0). */
    MeVector4 R1;
    /** Rotation matrix row 2 (last element 0). */
    MeVector4 R2;
    /** Position vector (last element 1). */
    MeVector4 pos;
} MdtKeaTransformation;

/**
 * Kea Body structure.
 *
 * After a time step, the contents of 'force' and 'torque' will be the
 * total applied forces plus any forces and torques due to constraints.
 * These MUST BE RE-ZEROED before accumulating values for the next time
 * step.
*/
typedef struct
{
    /** Data stucture tag. Should be MdtKeaBODYVER1. */
    int tag;                    
    /** Total length of data structure. */
    int len;

    /** 1/Mass of the body. */
    MeReal invmass;

    /**
     * Bitwise combination of MdtKeaBodyFlag's.
     * @see MdtKeaBodyFlag.
     */
    int flags;

    /** External force and torque (at centres of mass). */
    MeLinAng externalForce;            

    /** Resultant force and torque 
     *  (external force + constraint force at centres of mass). 
     */
    MeLinAng resultantForce;            

    /* ToDo: This should become an MeMatrix3. */
    /** Inverse of 3x3 inertia tensor - row 0. */
    MeVector4 invI0;            
    /** Inverse of 3x3 inertia tensor - row 1.  */
    MeVector4 invI1;            
    /** Inverse of 3x3 inertia tensor - row 2. */
    MeVector4 invI2;            

    /* ToDo: This should become an MeMatrix3. */
    /** The 3x3 inertia tensor - row 0.  */
    MeVector4 I0;               
    /** The 3x3 inertia tensor - row 1. */
    MeVector4 I1;               
    /** The 3x3 inertia tensor - row 2. */
    MeVector4 I2;               

    /** Linear velocity of body. */
    MeVector4 vel;              
    /** Angular velocity of body. */
    MeVector4 velrot;           
    /** Body orientation quaternion. */
    MeVector4 qrot;             

    /** Linearl Acceleration of body (READ ONLY). */
    MeVector4 accel;            
    /** Angular acceleration of body (READ ONLY). */
    MeVector4 accelrot;         

    /**
     * Axis of assumed fast rotation (unit length).
     * Only used if MdtKeaBodyFlagUseFastSpin is set in 'MdtKeaBody.flags'.
     */
    MeVector4 fastSpinAxis;     
}
MdtKeaBody;


/**
 * Sparse representation of velocity Jacobian matrix.
 */
typedef struct 
{
    MeReal linear[3];
    int    body;
    MeReal angular[3];
    int    constraint;
} MdtVelocityJacobianRowBody;

typedef struct 
{
    MdtVelocityJacobianRowBody body[2];
} MdtVelocityJacobianRow;

typedef struct 
{
    MdtVelocityJacobianRow *  row;
    int                       numRows;
    int                       numBodies;
    int                       maxRows;
} MdtVelocityJacobian;

/**
 * Constraints
 */
typedef struct {
    MdtVelocityJacobian     J;
    MeVector4N              xgamma;
    MeVector4N              c;
    MeVector4N              xi;
    MeVector4N              lo;
    MeVector4N              hi;
    MeVector4N              slipfactor;

    /** 
     * bodiesConst.elt[i] = pair of bodies constrained by constraint i 
     *
     * bodiesConst.elt[i].elt[0] is primary body constrained by constraint i
     * bodiesConst.elt[i].elt[1] is secondary body constrained by constraint i
     */
    MeI32PairArray          bodiesConstraint;      

    /* numRowsConst.elt[i] = number of rows in constraint i */
    MeI32Array              numRowsConstraint;

    /* numConstraintsPartition.elt[i] = number of constraints in partition i */
    MeI32Array              numConstraintsPartition;

    /* numRowsPartition.elt[i] = number of rows in partition i (excluding kea or arthur padding) */
    MeI32Array              numRowsPartition;

    /* numArthurRowsPartition.elt[i] = number of rows in partition i (including arthur padding at end of each constraint) */
    MeI32Array              numArthurRowsPartition;

    /* numBodiesPartition.elt[i] = number of bodies in partition */

    MeI32Array              numBodiesPartition;

    int                     numBodies;
    int                     numConstraints;
    int                     numPartitions;

    /* num rows padded only at end of partitions */
    int                     numRows;

    /* num rows padded at end of every constraint */
    int                     numArthurRows;

} MdtConstraints;

typedef enum
{
    MdtSolverSpecifyMaxPerFrame,
    MdtSolverSpecifyMaxPerPartition

} MdtSolverIterationCappingMethod;

typedef struct
{
    /** Amount of time to evolve system by. */
    MeReal stepsize;

    /** Numerical tolerance used by matrix solver. Default = 0.01. */
    MeReal epsilon;

    /** Constraint relaxation rate. Default = 0.2. */
    MeReal gamma;

    /** Maximum allowed number of LCP iterations. Default = 10 */
    int keaMaxIterations;


    /** Specifies whether we want to set the max number of iterations
     *  per partition, or maximum iterations per frame.
     *
     *  If set to MdtSolverSpecifyMaxPerFrame, then arthur
     *  does a total of arthurMaxIterationsTotal over all the partitions.
     *  The larger a partition, the larger the share of arthurMaxIterationsTotal
     *  it is allocated.
     *
     *  If set to MdtSolverSpecifyMaxPerPartition, then arthur
     *  does arthurMaxIterationsPerPartition iterations on each partition.
     */
    MdtSolverIterationCappingMethod arthurIterationCappingMethod;

    int arthurMaxIterationsTotal;
    int arthurMaxIterationsPerPartition;
    MeReal arthurVelocityZeroTol;
    MeReal arthurSafetyZeroTol;

    /** Velocity below which LCP considers velocity to be zero.
     *  Default = 0.03. 
     */
    MeReal keaVelocityZeroTol;

    /**
     * Pointer to some memory that Kea can use.
     *
     * Kea doesn't malloc, only uses this area.
     */
    void *memory_pool;

    /** Size of this area in bytes. */
    unsigned int memory_pool_size;

    /** CPU resources */
    MeCPUResources cpuResources;
}
MdtSolverParameters;

typedef struct {
    
    /* Constraint data */
    MdtConstraints *        constraints;
    
    /* Body data */
    MeLinAngPtrArray        externalForce;
    MeLinAngArray           unconstVel;
    MeMassMatrixArray       invMworld;

    /* Parameters (timestep etc) */
    MdtSolverParameters *   parameters;

} MdtSolverInputData;

typedef struct {
    MeVector4N              lambda;
    MeLinAngPtrArray        bodyForce;
    MeLinAngPairPtrArray    constraintForce;
} MdtSolverOutputData;

#ifdef __cplusplus
extern "C"
{
#endif

void MdtVelocityJacobianPrintTrans(const MdtVelocityJacobian *  J,     /* Input */
                                   const char *                 desc); /* Input */

void MdtVelocityJacobianPrint(const MdtVelocityJacobian *  J,     /* Input */
                              const char *                 desc); /* Input */

void MdtConstraintsPrint(const MdtConstraints *  c, /* Input */    
                         const char * desc);        /* Input */

void MEAPI MdtSolverUtilsCalcInvMworldAndUnconstVel(
        MeMassMatrixArray          invIworld,         /* Output */
		MeLinAngArray              unconstVel,        /* Output */
        MdtKeaBody *const          blist[],           /* Output/Input */
        const MdtKeaTransformation tlist[],           /* Input */
        int                        numBodies,         /* Input */
        MeReal                     stepsize);         /* Input */

void MEAPI MdtSolverUtilsCalculateAcceleration(
    MdtKeaBody *const             blist[],    /* input/output */
    const MeMassMatrixArray       invMworld,  /* input */
    int                           numBodies); /* input */

void MEAPI MdtSolverUtilsIntegrateSystem(MdtKeaBody *const       blist[],
                                 MdtKeaTransformation *  tlist,
                                 int                     num_bodies,
                                 MdtSolverParameters     parameters);

#ifdef __cplusplus
}
#endif

#endif
