#ifndef _MCDSPACE_H /* -*- mode: C; -*- */
#define _MCDSPACE_H

/*
   Copyright (c) 1997-2002 MathEngine PLC

   $Name:  $

   Date: $Date: 2003/05/05 00:57:07 $ - Revision: $Revision: 1.1 $

   This software and its accompanying manuals have been developed
   by MathEngine PLC ("MathEngine") and the copyright and all other
   intellectual property rights in them belong to MathEngine. All
   rights conferred by law (including rights under international
   copyright conventions) are reserved to MathEngine. This software
   may also incorporate information which is confidential to
   MathEngine.

   Save to the extent permitted by law, or as otherwise expressly
   permitted by MathEngine, this software and the manuals must not
   be copied (in whole or in part), re-arranged, altered or adapted
   in any way without the prior written consent of the Company. In
   addition, the information contained in the software may not be
   disseminated without the prior written consent of MathEngine.
*/

/**
 * @file
 *
 * Manages the large-scale spatial properties of a region of 3D space
 * populated by McdModel objects.
 *
 * Its principal role is to keep track of proximities between many
 * McdModel objects.
 *
 * Pairwise proximities are reported using a retained-mode mechanism
 * that maintains identity of a given pair over successive states: When
 * pairs of McdModel objects are first detected to be in close proximity,
 * a McdModelPair object is assigned to them.
 *
 * In subsequent steps, the same McdModelPair object will be re-used to
 * refer to the same pair.  When the pair is no longer in proximity, the
 * McdModelPair object is recycled as needed for use as a reference to a
 * differnt pair.
 *
 * The continuity of McdModelPair identity is exploited by many of the
 * Mcd interaction algorithms, and also by response modules such as
 * MathEngine's Dynamics Toolkit and Simulation Toolkit.
 */

#include <McdCTypes.h>
#include <McdCullingTable.h>
#include <McdModelPairManager.h>

#ifdef __cplusplus
extern "C" {
#endif

/*@{*/
/**
 *  McdSpace
 */
MEPUBLIC extern int McdXAxis;    /**< X-axis to be or'ed with McdYAxis and McdZAxis */
MEPUBLIC extern int McdYAxis;    /**< Y-axis to be or'ed with McdXAxis and McdZAxis */
MEPUBLIC extern int McdZAxis;    /**< Z-axis to be or'ed with McdXAxis and McdYAxis */
MEPUBLIC extern int McdAllAxes;  /**< A constant equal to McdXAxis+McdYAxis+McdZAxis.*/
/*@}*/

/*@{*/
/** Bounding volume update fns. */
typedef void (MEAPI *McdSpaceUpdateAABBFnPtr) (McdModelID model,
                                                MeReal duration);
typedef MeBool (MEAPI *McdSpaceUpdateBSphereFnPtr) (McdModelID model,
                                                    MeVector3 min,
                                                    MeVector3 max);
/*@}*/
/**
 *  Creates a new McdSpace object whose models are sorted by x, y, and z axes.
 *  @arg axes is a bit field, where the bits McdXAxis, McdYAxis,
 *    and McdZAxis may be used in combination to specify which axes
 *    to test for model proximity.  Axes left unspecified will cause all
 *    pairs to be treated as if they always overlap along those axes.  Note:
 *    McdAllAxes is a constant equal to McdXAxis+McdYAxis+McdZAxis.
 *  @arg objectCount is the maximum number of objects the space may hold.
 *  @arg pairCount is the maximum number of overlapping pairs the space
 *    will handle.
 *  @warning Performs memory allocation.
 *  @see McdXAxis McdYAxis McdZAxis McdAllAxes
 */
MEPUBLIC
McdSpaceID        MEAPI McdSpaceAxisSortCreate(McdFrameworkID fwk,
                            int axes,int objectCount,int pairCount);

/** Destroys @a s.
 * Does not destroy the McdModel objects present in @a s. They
 * can be destroyed manually using McdSpaceModelIterator, McdSpaceRemove and
 * McdModelDestroy. It is not required to remove all models from @a s before
 * calling this function.
 */
MEPUBLIC
void              MEAPI McdSpaceDestroy(McdSpaceID s);

/**
 * Inserts @a m into @a s. @a s will now keep track of the volume occupied by @a
 * m and detect any close proximities with other models present in @a s.
 * @m's bounding volume will be computed via McdModelGetAABB.
 * Note that McdSpaceInsertModel does not imply McdSpaceUpdateModel. You must
 * call the latter function explicitly ( or via McdSpaceUpdateAll ).
 * Once @m is updated, the next call to McdSpaceGetPairs will report any
 * close proximities between @a m and other McdModel objects present in @a
 * s. 

 * a McdModel object can only be present in one McdSpace at a time.

 * The hint parameter can be used to perform this operation more efficiently.
 * If the hint is a valid model in @a s, then the space's representation of
 * the model cm will initially be copied from that of the hint model.  If
 * the model's boundaries in space are close to those of the hint model's,
 * then the initial update (via McdSpaceUpdateModel or McdSpaceUpdateAll)
 * will in general be much faster than if the model was inserted using
 * McdSpaceInsertModel.  This operation is still Order(N), where N is the
 * number of models presently in the space, but the prefactor should be smaller.
 */
MEPUBLIC
MeBool            MEAPI McdSpaceInsertModel(McdSpaceID s,
                            McdModelID cm, McdModelID hint);

/** 
 * if already present in s, insert operation will be ignored,
 * in debug, a warning will also be issued.
 * zero return value indicates failure.. ? what kind ?, is failure
 * checked for in release ?
 * two models are culled immediately from the list of AABB-intersecting
 * pairs if they have the same culling table and culling ID, and the
 * bit indexed by their two cullingIndex values in the table is set.
 * @see McdSpaceInsertModel for a description of the hint parameter.
 */
MEPUBLIC
MeBool            MEAPI McdSpaceInsertModelWithCulling(McdSpaceID s, 
                            McdModelID cm, McdModelID hint,
                            McdCullingTable *table, int cullingIndex,
                            int cullingID);

/** 
 * Takes model out of its space and removes all model pairs.
 * @return 1 if successful, 0 if not.
 * 
 * @see McdSpaceUpdateAll()
 */
MEPUBLIC
MeBool            MEAPI McdSpaceRemoveModel(McdModelID);

/**
 * Computes overlaps with this model in its space.
 */
MEPUBLIC
void              MEAPI McdSpaceUpdateModel(McdModelID);

/** 
 * sets the culling parameters for the model.
 * @see McdSpaceInsertModelWithCulling 
 */
MEPUBLIC
void              MEAPI McdSpaceSetModelCullingParameters(McdSpaceID s, 
                            McdModelID cm,McdCullingTable *table,
                            int cullingIndex,int cullingID);

/**
 * Freeze the current McdSpace bounding volume data associated with @a m.
 * Subsequent calls to McdSpaceUpdateAllModels(),
 * will ignore @a m. Subsequent direct calls to McdSpaceUpdateModel()
 * using @a m as an argument will have no effect, and a warning will be
 * issued in debug mode.
 *
 *
 * Note that McdSpaceFreezeModel() does not imply a call to
 * McdSpaceUpdateModel(). If @a m's transform has changed since the last time
 * McdSpaceUpdateModel() was called on it, you may want to update @a m before
 * freezing it.
 * 
 * Pairs of models in which both pairs are frozen will continue to be
 * reported.
 * If @a m is a newly inserted model, be sure that it has been explicitly
 * updated ( it is not implicitly updated via McdSpaceInsertModel() ) before
 * freezingit.

 * After calling McdSpaceFreezeModel(m), subsequent calls to 
 * McdSpaceModelIsFrozen(m) will return true. @a m's "frozen status" can be
 * removed via McdSpaceUnFreezeModel().
 *
 * @see McdSpaceUnFreezeModel, McdSpaceUpdateModel, McdSpaceModelIsFrozen,
 * McdSpaceGetPairs
 *
 *  @warning You should not change the transform of a model (or associate it
 *  with a dynamics body while it is frozen

*/
MEPUBLIC
MeBool            MEAPI McdSpaceFreezeModel(McdModelID);


/**
 * Reverses effect of McdSpaceFreezeModel().
 * Subsequent calls to McdSpaceIsFrozen() using @a m will return false.
 * McdSpaceUpdateAllModels() will now update @a m as well.
 * 
 * If you are not using McdSpaceUpdateAllModels(), call McdSpaceUpdateModel() explicitly, 
 * if an update is required. ( McdSpaceUnfreezeModel() does not imply a call to
 * McdSpaceUpdateModel() )
 * 
 * Subsequent calls to McdSpaceModelIsFrozen with @a m as argument will
 * return false.
 * 
 * @see McdSpaceFreezeModel, McdSpaceUpdateAll, McdSpaceUpdateModel
*/
MEPUBLIC
MeBool            MEAPI McdSpaceUnfreezeModel(McdModelID);

/**
 * Returns @a cm 's "frozen" status.
 * @see McdSpaceFreezeModel, McdSpaceUnfreezeModel
 */
MEPUBLIC
MeBool            MEAPI McdSpaceModelIsFrozen(McdModelID);

/**
 * Update all unfrozen McdModel objects in @a s.
 * Calls McdSpaceUpdateModel() on all non-frozen models in @a s.
 */
MEPUBLIC
void              MEAPI McdSpaceUpdateAll(McdSpaceID s);

/**
 * Same effect as McdSpaceUpdateAll(), except that the usual AABB bounding
 * volumes for each McdModel object in @a s are replaced by PathAABB
 * bounding volumes.
 *
 * This function is subject to change.
 * 
 * @see McdModelGetPathAABB, McdSpaceUpdateAll
 */
MEPUBLIC
void              MEAPI McdSpacePathUpdateAll(McdSpaceID s, MeReal duration);

/** McdSpaceIsChanging */
MEPUBLIC
int               MEAPI McdSpaceIsChanging(McdSpaceID space);

/** McdSpaceSetAABBFn */
MEPUBLIC
void              MEAPI McdSpaceSetAABBFn(McdSpaceID s,
                            McdSpaceUpdateAABBFnPtr updateAABBFn);

/**
 * Intersects an oriented line segment with all models in the space.
 * @arg space the collision space
 * @arg inOrig pointer to an MeVector3 representing the first point on
 * the line segment.
 * @arg inDest pointer to an MeVector3 representing the second point on
 * the line segment.
 * @arg outList structure containing the line segment intersection data.
 * @arg inMaxListSize the maximum number of intersection which will be reported.
 * @see McdLineSegIntersect
*/
MEPUBLIC
int               MEAPI McdSpaceGetLineSegIntersections(McdSpaceID space,
                            MeReal *inOrig,MeReal *inDest,
                            McdLineSegIntersectResult *outList,
                            int inMaxListSize);
/**
 * Finds first intersection of an oriented line segment with all models
 * in the space.
 * @arg space the collision space
 * @arg inOrig pointer to an MeVector3 representing the first point on
 * the line segment.
 * @arg inDest pointer to an MeVector3 representing the second point on the
 * line segment.
 * @arg outList structure containing the line segment intersection data.
 * @see McdLineSegIntersect
 */
MEPUBLIC
int               MEAPI McdSpaceGetLineSegFirstIntersection(McdSpaceID space,
                            MeReal *inOrig, MeReal *inDest,
                            McdLineSegIntersectResult *outResult);
/** SetPoolFullHandler */
MEPUBLIC
void              MEAPI McdSpaceSetPoolFullHandler(McdSpaceID space,
                            McdSpacePoolErrorFnPtr handler);
/** GetPoolFullHandler */
MEPUBLIC
McdSpacePoolErrorFnPtr MEAPI McdSpaceGetPoolFullHandler(McdSpaceID space);

/**
 * Callback allowing the user to specify whether intersection with a
 * linesegment with a specific model is enabled.
 *
 * Return 0 if disabled, non-zero otherwise. filterData is passed.
 */
typedef int     (MEAPI *McdLineSegIntersectEnableCallback)(McdModelID,
                            void *filterData);

/**
 * Finds first intersection of an oriented line segment with all
 * enabled models in the space.
 * @arg space the collision space
 * @arg inOrig pointer to an MeVector3 representing the first point on
 * the line segment.
 * @arg inDest pointer to an MeVector3 representing the second point on
 * the line segment.
 * @arg outList structure containing the line segment intersection data.
 * @see McdLineSegIntersect
 */
MEPUBLIC
int               MEAPI McdSpaceGetLineSegFirstEnabledIntersection(
                            McdSpaceID space,MeReal *inOrig,MeReal *inDest,
                            McdLineSegIntersectEnableCallback filterCB,
                            void *filterData,
                            McdLineSegIntersectResult *outResult);

/**
 * Resume tracking of proximity between m1 and m2.
 * If m1 and m2 come into close proximity after this point,
 * or are in close proximity at the time that the call is made,
 * a new McdPair object will be assigned to this pair.
 *
 * Subsequent calls to McdSpacePairIsEnabled() with @a m1 and @a m2 as
 * arguments will return true.
 */

MEPUBLIC
MeBool            MEAPI McdSpaceEnablePair(McdModelID cm1, McdModelID cm2);
/**
 * Cease to track proximity between m1 and m2.
 * If m1 and m2 come into close proximity after this point,
 * no McdPair object will be assigned to this pair, and
 * nothing will be reported about the pair in the next call to
 * McdSpaceGetPairs().
 
 * If McdSpaceEnablePair() is subsequently called on
 * the same pair of models, the pair will reappear as a new McdPair
 * object, if the pair is in close proximity to each other. 

 * This behaviour is different than the one caused by a pair becoming frozen
 * via McdSpaceFreezeModel(). In that case, the frozen pair does not disappear.

 * Subsequent calls to McdSpacePairIsEnabled() with @a m1 and @a m2 as
 * arguments will return false.
 */
MEPUBLIC
MeBool            MEAPI McdSpaceDisablePair(McdModelID cm1, McdModelID cm2);

/**
 * Returns the "enabled" status for the pair @a cm1, @a cm2.
 * By default, pairs are enabled unless McdSpaceDisablePair() has been called.
 * @see McdSpaceEnablePair(), McdSpaceDisablePair().
*/
MEPUBLIC
MeBool            MEAPI McdSpacePairIsEnabled(McdModelID cm1, McdModelID cm2);

/**
 * Set user void* attached to space.
 */
MEPUBLIC
void              MEAPI McdSpaceSetUserData(McdSpaceID s, void *data);
/**
 *  Get user void* attached to space.
 */
MEPUBLIC
void *            MEAPI McdSpaceGetUserData(McdSpaceID s);

typedef McdModelPairManagerPairIterator McdSpacePairIterator;

#define McdSpacePairIteratorInit(_a,_b) McdModelPairManagerPairIteratorInit((_b));
#define McdSpacePairIteratorTerm(_a,_b) McdModelPairManagerPairIteratorTerm((_b));


/**
 * get all pair-events since the last call to McdSpaceUpdateAll(). By
 * default.  Return value
 * indicates overflow condition: call again to get remaining pairs.
*/
MEPUBLIC
int               MEAPI McdSpaceGetPairs(McdSpaceID s,
                                         McdSpacePairIterator* iter,
                                         McdModelPairContainer *);

/** Structure holding iteration information for McdSpaceGetModel. */
typedef struct
{
  int it;    /**< it */
  int count; /**< count */
} McdSpaceModelIterator;


/**
 * Initialise an iterator to iterate overall models in space.
 * Use McdSpaceGetModel to get successive models.
 */
MEPUBLIC
void              MEAPI McdSpaceModelIteratorInit(McdSpaceID,
                            McdSpaceModelIterator *it);

/**
 * Finalize a model iterator. Has no actual effect.
 */
MEPUBLIC
void              MEAPI McdSpaceModelIteratorTerm(McdSpaceID,
                            McdSpaceModelIterator *it);

/**
 * Get the model pointed to by iterator it, and advance iterator.
 * Returns 1 if the returned model is valid, 0 otherwise.
*/
MEPUBLIC
MeBool            MEAPI McdSpaceGetModel(McdSpaceID,
                            McdSpaceModelIterator *it,McdModelID *);

/**
 * Returns the number of models in Space
 */
MEPUBLIC
int               MEAPI McdSpaceGetModelCount(McdSpaceID space);

/**
 * returns the number of new pairs, and sets the pairArrayPtr
 * to point to them.
 */
MEPUBLIC          
int               MEAPI McdSpaceGetNewPairs(McdSpaceID space, McdModelPairID **pairListPtr);

#ifdef __cplusplus
}
#endif


#endif
